/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Preferences Toolbar 4.
 *
 * The Initial Developer of the Original Code is Manuel Reimer.
 *
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Aaron Andersen <aaron@xulplanet.com>
 *                 Kevin Teuscher
 *                 Manuel Reimer <Manuel.Reimer@gmx.de>
 *                 Stephen Clavering <mozilla@clav.me.uk>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var prefbarContextShiftDown = false;

// goPrefBar sets this to true, if anything in database has changed
var prefbarDatabaseChanged = false;

var prefbarToolbarsCustomizing = false;

var chevron_enabled = true;

window.addEventListener("load", prefbarStartPrefBar, true);

function prefbarStartPrefBar(event) {
    window.removeEventListener("load", prefbarStartPrefBar, true);

    var buttons = document.getElementById("prefbar-buttons");
    var chevron = document.getElementById("prefbar-chevron-popup");

    goPrefBar.RDF.AddmDatasource(buttons);
    // Call init functions as early as possible, to be sure, that
    // buttons are able to hook up Observers or stuff like that before the
    // first page gets loaded.
    prefbarCallInitFunctions();
    goPrefBar.RDF.AddmDatasource(chevron);

    // HACK On Mac OS X the menubar seems to be cached somewhere
    // https://www.mozdev.org/bugs/show_bug.cgi?id=18068
    var pbmenu = document.getElementById("prefbar-menu");
    if (pbmenu) pbmenu.parentNode.removeChild(pbmenu);

    var menubar = document.getElementById("main-menubar");
    if (menubar) {
        var prefbarmenu = document.createElement('menu');
        var prefbarpopup = document.createElement('menupopup');
        prefbarmenu.id = "prefbar-menu";
        prefbarmenu.setAttribute("label", "PrefBar");
        prefbarpopup.id = "prefbar-menu-popup";
        prefbarpopup.setAttribute("datasources", "rdf:null");
        prefbarpopup.setAttribute("template", "prefbarToolbarTemplate");
        prefbarpopup.setAttribute("ref", "urn:prefbar:browserbuttons:enabled");
        prefbarpopup.setAttribute("flags", "dont-test-empty");
        prefbarpopup.setAttribute("onpopuphidden", "setTimeout('prefbarSetChecks(false, false)');"); // setTimeout required to keep toolbar and menu in sync
        prefbarpopup.setAttribute("onpopupshowing", "prefbarSetChecks(true, false, this);");
        menubar.appendChild(prefbarmenu);
        prefbarmenu.appendChild(prefbarpopup);
        goPrefBar.RDF.AddmDatasource(prefbarpopup);
        // Set "collapsed" attribute. This, for any reason, has to be set
        // *after* all Elements are linked together.
        prefbarmenu.setAttribute("collapsed", Boolean(!navigator.preference("extensions.prefbar.show_prefbar_menu")));
    }

    // The "Preferences Toolbar" menuitem is placed directly under "View", by
    // default, on Firefox. Try to move it to View->Toolbars
    var toolbarsmenu = document.getElementById("viewToolbarsMenu");
    var viewpbmenu = document.getElementById("viewprefsbar");
    if (toolbarsmenu && viewpbmenu) {
        toolbarsmenu = toolbarsmenu.childNodes[0];
        var newpos = toolbarsmenu.childNodes.length - 2;
        viewpbmenu.parentNode.removeChild(viewpbmenu);
        toolbarsmenu.insertBefore(viewpbmenu, toolbarsmenu.childNodes[newpos]);
    }

    // If we are not allowed to display in Popups, then bind PrefBar to
    // the class "chromeclass-toolbar" to let the browser-backend hide it on
    // popups.
    if (!navigator.preference("extensions.prefbar.show_in_popups")) {
        var prefbar = document.getElementById("prefbar");
        prefbar.setAttribute("class", "chromeclass-toolbar");
    }

    // Register a few event listeners
    window.addEventListener("focus", prefbarFocus, false);
    window.addEventListener("resize", prefbarResize, false);

    var appcontent = window.document.getElementById("appcontent");
    appcontent.addEventListener("click", prefbarLinkClicked, false);
    appcontent.addEventListener("DOMContentLoaded", prefbarPageLoaded, false);
    appcontent.addEventListener("select", prefbarTabChanged, false);

    // DragDrop listener, required to detect, if firefox toolbars are currently
    // customized
    if (goPrefBar.InFF) {
        var toolbox = document.getElementById("navigator-toolbox");
        toolbox.addEventListener("dragdrop", prefbarToolboxDND, false);
        toolbox.addEventListener("draggesture", prefbarToolboxDND, false);
    }

    // Bugfix for Firefox. An old workaround to toggle the toolbar forces us
    // to add the following lines
    if (goPrefBar.InFF) {
        var toolbar = document.getElementById("prefbar");
        toolbar.setAttribute("collapsed", false);
    }

    // MultiZilla support. Redraw toolbar whenever the QPrefs menu has been
    // used
    var mztabcontextmnu = document.getElementById("tabContextMenu");
    var mzquickprefsmnu = document.getElementById("quickpref-popup");
    if (mzquickprefsmnu && mztabcontextmnu) {
        goPrefBar.dump("prefbarStartPrefbar: Registering Event Listeners for MultiZilla QPrefs");
        mztabcontextmnu.addEventListener("popuphidden", function(){setTimeout('prefbarSetChecks(false, false)');}, false);
        mzquickprefsmnu.addEventListener("popuphidden", function(){setTimeout('prefbarSetChecks(false, false)');}, false);
    }

    // Init hotkey stuff
    prefbarDoHotkeySnapshot();
    prefbarUpdateToggleKey();
    prefbarUpdateButtonHotkeys();

    // Set value for "slimbuttons"
    var slimvalue = navigator.preference("extensions.prefbar.slimbuttons");
    buttons.setAttribute("prefbarslimbuttons", slimvalue);

    // Move PrefBar, if needed. This also calls prefbarUpdateToolbar!
    // We do this asynchronously to speed up browser startup and to get sure,
    // that "chromehidden" is already set, if "prefbarMoveButtons" gets called
    setTimeout('prefbarMoveButtons();', 0);

    prefbarUpdateNotify();
}

// preferences observer "driven" by goPrefBar
function prefbarPrefObserver(subject, topic, data) {
    var value;
    switch (data) {
    case "extensions.prefbar.show_prefbar_menu":
        value = navigator.preference(data);
        var pbmenu = document.getElementById("prefbar-menu");
        pbmenu.setAttribute("collapsed", !value);
        break;
    case "extensions.prefbar.display_on":
        prefbarMoveButtons();
        break;
    case "extensions.prefbar.slimbuttons":
        value = navigator.preference(data);
        var buttons = document.getElementById("prefbar-buttons");
        buttons.setAttribute("prefbarslimbuttons", value);
        break;
    case "extensions.prefbar.hktoggle":
        prefbarUpdateToggleKey();
        break;
    }
}

var prefbarLastWidth = 0;
var prefbarLastHeight = 0;
function prefbarResize(event) {
    if (event.eventPhase != event.AT_TARGET) return;

    if (prefbarLastWidth == window.innerWidth &&
        prefbarLastHeight == window.innerHeight) {
        goPrefBar.dump("prefbarResize: ignored");
        return;
    }
    prefbarLastWidth = window.innerWidth;
    prefbarLastHeight = window.innerHeight;

    goPrefBar.dump("prefbarResize");
    setTimeout(prefbarUpdateToolbar);
}

function prefbarFocus(event) {
//    if (event.target != document ||
//        event.eventPhase != event.AT_TARGET)

//    if (event.target != document)
//        return;

    if (event.eventPhase != 2) return;

    goPrefBar.dump("prefbarFocus");

    if (prefbarDatabaseChanged) {
        prefbarDatabaseChanged = false;
        // Update the hotkeys first, to be sure the right keys
        // are displayed in menus
        prefbarUpdateButtonHotkeys();
        // Now update anything else...
        document.getElementById("prefbar-buttons").builder.rebuild();
        prefbarCallInitFunctions();
        document.getElementById("prefbar-chevron-popup").builder.rebuild();
        document.getElementById("prefbar-menu-popup").builder.rebuild();
        setTimeout(prefbarUpdateToolbar);
    }
    else
        prefbarSetChecks(false, false);

    // Check if any of the toolkit toolbars have been customized.
    var cmdToolbarCustomize = document.getElementById("cmd_CustomizeToolbars");
    if (cmdToolbarCustomize) {
        var cmd_disabled = cmdToolbarCustomize.getAttribute("disabled");
        if (!cmd_disabled && prefbarToolbarsCustomizing) {
                prefbarToolbarsCustomizing = false;
                prefbarMoveButtons();
        }
    }
}

// DragDrop listener, required to detect if toolkit toolbars are currently
// customized
function prefbarToolboxDND(event) {
    var cmdToolbarCustomize = document.getElementById("cmd_CustomizeToolbars");
    if (cmdToolbarCustomize) {
        var cmd_disabled = cmdToolbarCustomize.getAttribute("disabled");
        if (cmd_disabled == "true")
            prefbarToolbarsCustomizing = true;
    }
}

function prefbarLinkClicked(event) {
    var node = event.originalTarget;
    if (!node.getAttribute) return true;
    if (!node.hasAttribute("href")) return true;
    var tagname = node.nodeName.toLowerCase();
    var href = node.getAttribute("href");
    if (tagname == "a" && href.indexOf("prefbar://") == 0 && event.button < 2) {
        goPrefBar.dump("PrefBar: Link click " + href);
        window.openDialog("chrome://prefbar/content/urlimport.xul", "prefbarURLImport", "chrome,centerscreen,modal,titlebar", href);
        event.stopPropagation();
        event.preventDefault();
        return false;
    }
    return true;
}

function prefbarPageLoaded(event) {
    goPrefBar.dump("prefbarPageLoaded");
    prefbarSetSpecialChecks("page");
}

var prefbarLastTab = null;
function prefbarTabChanged(event) {
    var curtab = gBrowser.mCurrentTab;
    if (prefbarLastTab == curtab) {
        goPrefBar.dump("prefbarTabChanged: ignored");
        return;
    }
    goPrefBar.dump("prefbarTabChanged");
    prefbarLastTab = curtab;

    prefbarSetSpecialChecks("tab");
}

function prefbarSetSpecialChecks(updatefor) {
    var buttons = document.getElementById("prefbar-buttons");

    var len = buttons.childNodes.length;
    for (var i = 0; i < len; i++) {
        var button = buttons.childNodes[i];

        if (button.collapsed == true) break;

        var btnufor = button.getAttribute("browserbtnupdatefor");
        if (!btnufor) continue;
        if (updatefor == "page" && btnufor != "page") continue;

        var btntype = button.getAttribute('btntype');

        switch (btntype) {
        case "extcheck":
            prefbarSetExtcheck(button, false);
            break;
        case "extlist":
            prefbarSetExtlist(button.menupopup, false);
            break;
        }
    }
}


function prefbarDoHotkeySnapshot() {
    // already filled? Then don't waste CPU time again
    if (goPrefBar.hotkeysnapshot) return;

    goPrefBar.hotkeysnapshot = new Array();
    var keys = document.getElementsByTagName("key");
    for (var keyindex = 0; keyindex < keys.length; keyindex++) {
        var curkey = keys[keyindex];
        var modifier = curkey.getAttribute("modifiers");
        var key = curkey.getAttribute("key");
        if (key) key = key.toUpperCase();
        var keycode = curkey.getAttribute("keycode");
        if (keycode) keycode = keycode.toUpperCase();
        if (modifier) {
            var modarray = modifier.split(",");
            modarray.sort();
            for (var modindex = 0; modindex < modarray.length; modindex++) {
                var modkey = modarray[modindex]
                modkey = modkey.replace(/(^\s*)|(\s*$)/, "");
                if (modkey == "control") modkey = "accel";
                modarray[modindex] = modkey;
            }
            modifier = modarray.join(",");
        }
        var s = key + "\t" + keycode + "\t" + modifier;
        goPrefBar.hotkeysnapshot[s] = true;
    }
}

function prefbarUpdateToggleKey() {
    var mainwin = document.getElementById("main-window");
    var ourkeyset = document.getElementById("prefbarHKKeyset");

    if (ourkeyset)
        mainwin.removeChild(ourkeyset);

    var ourkeyset = document.createElement("keyset");
    ourkeyset.id = "prefbarHKKeyset";

    var hkstr = navigator.preference("extensions.prefbar.hktoggle");
    var hkarr = hkstr.split("][");
    if (hkarr.length != 3)
        return;

    // Only create the key, if the user enabled it (values are valid)
    if (hkarr[1] != "" || hkarr[2] != "") {
        var togglekey = document.createElement("key");
        togglekey.id= "key_prefbar";
        togglekey.setAttribute("modifiers", hkarr[0]);
        togglekey.setAttribute("key", hkarr[1]);
        togglekey.setAttribute("keycode", hkarr[2]);
        togglekey.setAttribute("command", "cmd_prefbar");
        ourkeyset.appendChild(togglekey);
    }

    mainwin.appendChild(ourkeyset);

    var menuitem = document.getElementById("viewprefsbar");
    menuitem.removeAttribute("acceltext");
}

function prefbarUpdateButtonHotkeys() {
    var mainwin = document.getElementById("main-window");
    var ourkeyset = document.getElementById("prefbarButtonKeyset");

    if (ourkeyset)
        mainwin.removeChild(ourkeyset);

    ourkeyset = document.createElement("keyset");
    ourkeyset.id = "prefbarButtonKeyset";

    ourkeyset.setAttribute("template", "prefbarHotkeyTemplate");
    ourkeyset.setAttribute("datasources", "rdf:null");
    ourkeyset.setAttribute("ref", "urn:prefbar:browserbuttons:enabled");
    ourkeyset.setAttribute("flags", "dont-test-empty");
    mainwin.appendChild(ourkeyset);

    goPrefBar.RDF.AddmDatasource(ourkeyset);
    ourkeyset.builder.rebuild();

    prefbarPrefixSubIds(ourkeyset, "key:");

    // keysets aren't properly initialised, when rendered using RDF rules
    var len = ourkeyset.childNodes.length;
    for (var cindex = 0; cindex < len; cindex++) {
        var curChild = ourkeyset.childNodes[cindex];
        if (curChild.tagName == "keyset") {
            var keyset = curChild.cloneNode(true);
            ourkeyset.insertBefore(keyset, curChild);
            ourkeyset.removeChild(curChild);
        }
    }
}
function prefbarPrefixSubIds(parent, prefix) {
    var len = parent.childNodes.length;
    for (var cindex = 0; cindex < len; cindex++) {
        var curChild = parent.childNodes[cindex];
        var curid = curChild.id;
        if (curid && curid.substr(0, prefix.length) != prefix)
            curChild.id = prefix + curid;
        if (curChild.childNodes.length > 0)
            prefbarPrefixSubIds(curChild, prefix);
    }
}


function prefbarMoveButtons() {
    var targetindex = navigator.preference("extensions.prefbar.display_on");

    // If we are asked to display somewhere else than "own toolbar"
    // and we are displayed on a popup window, where our target is hidden
    // then fall back to "own toolbar", so we aren't hidden
    var mainwin = document.getElementById("main-window");
    var chide = mainwin.getAttribute("chromehidden");
    if (targetindex == 1 && chide.match(/menubar/)) targetindex = 0;
    if (targetindex == 2 && chide.match(/(location|toolbar)/)) targetindex = 0;
    if (targetindex == 3 && chide.match(/directories/)) targetindex = 0;

    var menuitem = document.getElementById("viewprefsbar");
    var buttons = document.getElementById("prefbar-buttons");
    var contents = document.getElementById("prefbar-contents");

    var menubar = document.getElementById("main-menubar");
    var menubar_ff = document.getElementById("toolbar-menubar");

    var toolbars = [document.getElementById("prefbar"),
                    menubar_ff ? menubar_ff : menubar,
                    document.getElementById("nav-bar"),
                    document.getElementById("PersonalToolbar")];

    var target = toolbars[targetindex];

    if (targetindex == 0) {
        menuitem.collapsed = false;
        toolbars[0].collapsed = false;
    }
    else {
        menuitem.collapsed = true;
        toolbars[0].collapsed = true;
    }

    // Check if any of the toolbar-items are flexible. If so, then make
    // PrefBar *not* flexible to take up all free space
    chevron_enabled = true;
    contents.flex = "1";
    var len = target.childNodes.length;
    for (var cindex = 0; cindex < len; cindex++) {
        var curChild = target.childNodes[cindex];
        if (curChild.getAttribute("hidden") == "true") continue;
        if (curChild == contents) continue;
        if (curChild.tagName == "toolbarspring") continue;

        if (Number(curChild.flex) >= 1) {
            contents.flex = "0";
            chevron_enabled = false;
            break;
        }
    }

    var currentParent = contents.parentNode;
    if (currentParent != target) {
        currentParent.removeChild(contents);
        target.appendChild(contents);
        var chevron = document.getElementById("prefbar-chevron-popup");
        goPrefBar.RDF.AddmDatasource(buttons);
        goPrefBar.RDF.AddmDatasource(chevron);
        // Switched to "on own toolbar", but toolbar hidden
        // --> Show toolbar
        if (targetindex == 0 && target.hidden)
            prefbarTogglePrefBar()
    }

    prefbarUpdateToolbar();
    prefbarCallInitFunctions();

    if ("BookmarksToolbar" in this && "resizeFunc" in BookmarksToolbar)
        BookmarksToolbar.resizeFunc();
}


// This one will init all currently available buttons.
function prefbarCallInitFunctions(parent) {
    if (!parent) parent = document.getElementById("prefbar-buttons");

    var len = parent.childNodes.length;
    for (var cindex = 0; cindex < len; cindex++) {
        var button = parent.childNodes[cindex];
        var type = button.getAttribute("btntype");
        if (type == "submenu")
            prefbarCallInitFunctions(button.childNodes[0]);
        else {
            var initfunction = button.getAttribute("initfunction");
            if (initfunction) {
                try {
                    var lf = new Error();
                    eval(initfunction);
                } catch(e) {prefbarLogError(e, lf, button.id, 'initfunction');}
            }
        }
    }
}

function prefbarUpdateToolbar() {
    // Force to redraw everything as we need the button width.
    prefbarSetChecks(false, true);

    goPrefBar.dump("UpdateToolbar");

    // Now run the real update of the toolbar buttons asynchrously. This
    // is required, as the buttons need some time to finish drawing and we
    // need finished buttons for the following steps.
    setTimeout(prefbarDelayedUpdateToolbar);
}
function prefbarDelayedUpdateToolbar() {
    var buttons = document.getElementById("prefbar-buttons");
    var chevron = document.getElementById("prefbar-chevron");

    if (!buttons.firstChild) {
        // No buttons means no chevron
        chevron.collapsed = true;
        return;
    }

    var spacers = new Array()

    chevron.collapsed = false;
    var chevronWidth = chevron.boxObject.width;
    chevron.collapsed = true;

    // var toolbarWidth = window.innerWidth;  innerWidth == 0 while minimized!
    var toolbarWidth = buttons.boxObject.x + buttons.boxObject.width;
    var overflowed = false;

    var len = buttons.childNodes.length;
    for (var i = 0; i < len; i++) {
        var button = buttons.childNodes[i];
        if (chevron_enabled) {
            button.collapsed = overflowed;

            if (i == buttons.childNodes.length - 1)
                chevronWidth = 0;

            if (button.tagName == "spacer") {
                spacers.push(button);
                button.setAttribute("hidden", true);
                continue;
            }

            var remainingWidth = toolbarWidth - (button.boxObject.x + button.boxObject.width);
            if (remainingWidth < chevronWidth) {
                overflowed = true;
                // This button doesn't fit. Show it in the menu. Hide it in the
                // toolbar.
                if (!button.collapsed)
                    button.collapsed = true;
                if (chevron.collapsed) {
                    chevron.collapsed = false;
                }
            }
        }
        else {
            button.collapsed = false;
            chevron.collapsed = true;
        }
    }

    len = spacers.length;
    for (var s = 0; s < len; s++) {
        spacers[s].setAttribute("hidden", false);
    }
}


function prefbarUpdateMenu() {
    goPrefBar.dump("UpdateMenu");

    prefbarSetChecks(true, false);

    var toolbar = document.getElementById("prefbar-buttons");
    var menu = document.getElementById("prefbar-chevron-popup");
    var spacercount = 0;

    len = toolbar.childNodes.length;
    for (var i = 0; i < len; i++) {
        var button = toolbar.childNodes[i];
        if (button.tagName == "spacer") {
            spacercount++;
            continue;
        }
        var menuitem = menu.childNodes[i - spacercount];
        if (menuitem.hidden == button.collapsed)
            menuitem.hidden = !menuitem.hidden;
    }
}

function prefbarSetChecks(menu, forceAll, optional_target) {
    goPrefBar.dump("SetChecks Menu:" + !!menu);

    var buttons;

    if (optional_target != undefined)
        buttons = optional_target;
    else {
        if (!menu)
            buttons = document.getElementById("prefbar-buttons");
        else
            buttons = document.getElementById("prefbar-chevron-popup");
    }

    var len = buttons.childNodes.length;
    for (var i = 0; i < len; i++) {

        var button = buttons.childNodes[i];
        if (!forceAll && button.collapsed == true) {
            goPrefBar.dump("SetChecks: jumpingbutton");
            break;
        }
        var btntype = button.getAttribute('btntype');

        switch (btntype) {
        case "check":
            prefbarSetCheck(button, menu);
            break;
        case "extcheck":
            prefbarSetExtcheck(button, menu);
            break;
        case "menulist":
            if (!menu) prefbarSetMenulist(button.menupopup, false);
            break;
        case "extlist":
            if (!menu) prefbarSetExtlist(button.menupopup, false);
            break;
        }
    }
}

function prefbarProcessButton(button, event) {
    try {
        var lf = new Error();
        eval(button.getAttribute("oncommandval"));
    } catch(e) { prefbarLogError(e, lf, button.id, "onclick"); }
}

function prefbarSetCheck(button, menu) {
    // Value is magic variable referenced in prefstring
    var value = navigator.preference(button.getAttribute("prefstring"));

    var checked = eval(button.getAttribute("frompref"));

    if (!menu)
        button.setAttribute("checked", checked);
    else {
        var tvalue = checked ? "checkbox" : "radio";
        button.setAttribute("type", tvalue);
        button.setAttribute("checked", "true");
    }
}

function prefbarProcessCheck(button, menu) {
    var value;
    if (!menu)
        value = button.checked;
    else
        value = (button.getAttribute("type") == "radio");

    // Checkboxes keep focus, but we want the focus in the content area
    // https://www.mozdev.org/bugs/show_bug.cgi?id=18086
    window._content.focus();

    navigator.preference(button.getAttribute("prefstring"), eval(button.getAttribute("topref")));
}

function prefbarHotkeyCheck(key) {
    var value = navigator.preference(key.getAttribute("prefstring"));
    var checked = eval(key.getAttribute("frompref"));
    value = !checked;
    value = eval(key.getAttribute("topref"));
    navigator.preference(key.getAttribute("prefstring"), value);

    var msg;
    if (!checked)
        msg = goPrefBar.GetString("prefbarOverlayStrings", "enabled");
    else
        msg = goPrefBar.GetString("prefbarOverlayStrings", "disabled");
    prefbarOSDMessage(key.getAttribute("label") + ": " + msg);
    prefbarSetChecks(false, false);
}

function prefbarSetExtcheck(button, menu) {
    var func = button.getAttribute("getfunction");

    var value;
    try {
        var lf = new Error();
        eval(func);
    } catch(e) { prefbarLogError(e, lf, button.id, "getfuntion"); }

    if (!menu)
        button.setAttribute("checked", value);
    else {
        var tvalue = value ? "checkbox" : "radio";
        button.setAttribute("type", tvalue);
        button.setAttribute("checked", "true");
    }
}

function prefbarProcessExtcheck(button, menu, event) {
    var func = button.getAttribute("setfunction");

    var value;
    if (!menu)
        value = button.checked;
    else
        value = (button.getAttribute("type") == "radio");

    // Checkboxes keep focus, but we want the focus in the content area
    // https://www.mozdev.org/bugs/show_bug.cgi?id=18086
    window._content.focus();

    try {
        var lf = new Error();
        eval(func);
    } catch(e) { prefbarLogError(e, lf, button.id, "setfuntion"); }
}

function prefbarHotkeyExtcheck(key) {
    var value;
    eval(key.getAttribute("getfunction"));
    value = !value;
    eval(key.getAttribute("setfunction"));

    var msg;
    if (value)
        msg = goPrefBar.GetString("prefbarOverlayStrings", "enabled");
    else
        msg = goPrefBar.GetString("prefbarOverlayStrings", "disabled");
    prefbarOSDMessage(key.getAttribute("label") + ": " + msg);
    prefbarSetChecks(false, false);
}


function prefbarSetMenulist(menupopup, menu) {
    var prefstring = menupopup.getAttribute("prefstring");

    var prefBranch = goPrefBar.PrefBranch;
    var prefHasUserValue = prefBranch.prefHasUserValue(prefstring);

    var prefvalue;
    try {
        prefvalue = String(navigator.preference(prefstring));
    } catch(e) {}

    var items = prefbarGetMenupopupItems(menupopup);
    prefbarGenerateMenupopup(menupopup, menu, items, !prefHasUserValue, prefvalue, "prefbarProcessMenulist(this)");
}

function prefbarProcessMenulist(menuitem) {
    var prefBranch = goPrefBar.PrefBranch;

    var pref = menuitem.parentNode.getAttribute("prefstring");
    var value = menuitem.getAttribute("value");
    var type = prefBranch.getPrefType(pref);

    if (value == "PREFBARDEFAULT")
        prefBranch.clearUserPref(pref);
    else {
        if(type == prefBranch.PREF_STRING)
            prefBranch.setCharPref(pref, value);
        else if(type == prefBranch.PREF_INT)
            prefBranch. setIntPref(pref, Number(value));
        else if(type == prefBranch.PREF_BOOL)
            prefBranch.setBoolPref(pref, Boolean(value));
        else
            navigator.preference(pref, value);  // Couldn't hurt to try...
    }
}


function prefbarSetExtlist(menupopup, menu) {
    var getfunction = menupopup.getAttribute("getfunction");

    // The variables that may be modified by the getfunction
    /* value to select in list */
    var value = "";
    /* Here we have the array of all items defined via GUI. The array has
       two dimensions. First is the index and second is label(0) and value(1)
       feel free to edit,read,parse special values or dump new items in here
       from wherever you read them */
    var items = prefbarGetMenupopupItems(menupopup);

    var defaultset = false;

    // Call the getfunction
    try {
        var lf = new Error();
        eval(getfunction);
    } catch(e) { prefbarLogError(e, lf, menupopup.id, "getfuntion"); }

    prefbarGenerateMenupopup(menupopup, menu, items, defaultset, value, "prefbarProcessExtlist(this)");
}


function prefbarProcessExtlist(menuitem) {
    var func = menuitem.parentNode.getAttribute("setfunction");
    var value = menuitem.getAttribute("value");

    try {
        var lf = new Error();
        eval(func);
    } catch(e) { prefbarLogError(e, lf, menuitem.parentNode.id, "setfuntion"); }
}

function prefbarGetMenupopupItems(menupopup) {
    var RDF = goPrefBar.RDF.RDFService;
    var Datasource = goPrefBar.RDF.mDatasource;

    var itemId = menupopup.getAttribute("id");
    var RitemId = RDF.GetResource(itemId);
    var returnArray = Array();

    // It would be possible to implement some "cache" here
    // means a "hash array" which holds the finished arrays for each id.
    // Don't know if this is worth the effort and if this is really faster
    // than reading from RDF.

    var itemindex = 1;
    while(true) {
        var Roptlabel = RDF.GetResource(goPrefBar.RDF.NC + "optionlabel" + itemindex);
        var Noptlabel = Datasource.GetTarget(RitemId, Roptlabel, true);
        if (!Noptlabel) break;
        var optlabel = Noptlabel.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
        var Roptvalue = RDF.GetResource(goPrefBar.RDF.NC + "optionvalue" + itemindex);
        var Noptvalue = Datasource.GetTarget(RitemId, Roptvalue, true);
        //if (!Noptvalue) break;
        var optvalue = Noptvalue.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

        returnArray.push(Array(optlabel, optvalue));
        itemindex++;
    }

    return returnArray;
}

function prefbarGenerateMenupopup(menupopup, menu, items, defaultset, value, cmdhandler) {
    var itemId = menupopup.getAttribute("id");
    goPrefBar.dump("prefbarGenerateMenupopup: itemId: " + itemId);

    var activeitem = null;
    var defaultitem = null;

    while(menupopup.firstChild) menupopup.removeChild(menupopup.firstChild);

    var len = items.length;
    for (var itemindex = 0; itemindex < len; itemindex++) {
        var optlabel = items[itemindex][0];
        var optvalue = items[itemindex][1];

        var newitem = document.createElement('menuitem');
        if (optvalue == "PREFBARDEFAULT") defaultitem = newitem;
        if (optvalue == value) activeitem = newitem;

        newitem.setAttribute("label", optlabel);
        newitem.setAttribute("value", optvalue);
        newitem.setAttribute("oncommand", cmdhandler);
        if (menu) {
            newitem.setAttribute("type", "checkbox");
            newitem.setAttribute("checked", "false");
        }
        menupopup.appendChild(newitem);
    }

    if (defaultset && defaultitem) activeitem = defaultitem;

    if (menu)
        activeitem.setAttribute("checked", "true");
    else {
        if (activeitem)
            menupopup.parentNode.selectedItem = activeitem;
        else {
            menupopup.parentNode.selectedItem = null;
            var label = menupopup.parentNode.getAttribute("xlabel");
            menupopup.parentNode.setAttribute("label", label);
        }
    }
}


function prefbarOSDMessage(message) {
    var osd = document.getElementById("prefbarOSDMessage");
    osd.value = message;
    osd.collapsed = false;
    setTimeout('document.getElementById("prefbarOSDMessage").collapsed = true;', 1000);
}

function prefbarGoLink(url, event) {
    var pref_opentab = navigator.preference("browser.tabs.opentabfor.middleclick");
    var pref_openwindow = navigator.preference("middlemouse.openNewWindow");
    var pref_tabbackground = navigator.preference("browser.tabs.loadInBackground");
    var browser = window.document.getElementById("content");

    if (!event) {
        browser.loadURI(url);
        window.content.focus();
    }
    else {
        if (event.shiftKey) pref_tabbackground = !pref_tabbackground;

        var middleClick = true;
        if (event.type == "command" && !event.ctrlKey && !event.metaKey)
            middleClick = false;

        if (!middleClick) {
            browser.loadURI(url);
            window.content.focus();
        }
        else {
            if (pref_opentab) {
                var tab = browser.addTab(url);
                if (!pref_tabbackground)
                    browser.selectedTab = tab;
            } else if (pref_openwindow)
                window.open(url);
        }
    }
}

function prefbarTogglePrefBar() {
    var prefbar = document.getElementById("prefbar");
    if (prefbar.collapsed) return;
    var disppb = window.getComputedStyle(prefbar, "").getPropertyValue("display");
    // SeaMonkey hides the toolbar using the "hidden" attribute. Don't disable
    // the hotkey in this case!
    if (disppb == "none" && !prefbar.hidden) return;
    goToggleToolbar('prefbar','viewprefsbar');
    prefbarUpdateToolbar();
    window.content.focus();
}

function prefbarOpenPrefs() {
    var wm = Components.classes['@mozilla.org/appshell/window-mediator;1']
             .getService(Components.interfaces.nsIWindowMediator);

    var editWin = wm.getMostRecentWindow("prefbar:btneditor");
    if (editWin) {
        editWin.focus();
        return;
    }

    if (goPrefBar.InSM2) {
        var prefWin = wm.getMostRecentWindow("mozilla:preferences");
        if (prefWin) {
            var dialog = prefWin.document.getElementById("prefDialog");
            var pane = prefWin.document.getElementById("prefbar_editbar_pane");
            dialog.showPane(pane);
            prefWin.focus();
        }
        else
            goPreferences('prefbar_editbar_pane');
    }
    else if ("goPreferences" in this)
        goPreferences('navigator.xul', 'chrome://prefbar/content/preferencesPanel.xul', 'prefbarPanel');
    else {
        var prefWin = wm.getMostRecentWindow("prefbar:preferences");
        if (prefWin)
            prefWin.focus();
        else
            openDialog('chrome://prefbar/content/prefbar-bird-pref.xul', 'PrefBar', 'chrome');
    }
}

function prefbarOnContextMenu(event) {
    // Don't allow this uninstaller, if extension manager exists
    if ("@mozilla.org/extensions/manager;1" in Components.classes) {
        document.getElementById("prefbar-uninstall").setAttribute("hidden", "true");
        return;
    }

    // Only show the unistall item if the shift-key is down!
    //if (!event.shiftKey) // Once again an bug :-( We have to catch the status
    // of the shift-Key from the onmousedown and onkeydown events of the
    // "toolbar" item as it is always false here.
    if (!prefbarContextShiftDown)
        document.getElementById("prefbar-uninstall").setAttribute("hidden", "true");
    else
        document.getElementById("prefbar-uninstall").setAttribute("hidden", "false");
}

function prefbarUninstall() {
    if (!confirm(goPrefBar.GetString("prefbarOverlayStrings", "uninstallconfirm"))) return;

    var unreg = new prefbarExUnregisterer (
         'chrome://prefbar/content/contents.rdf',
         'jar:%chromeFolder%prefbar.jar!/locale/en-US/prefbar/contents.rdf',
         'jar:%chromeFolder%prefbar.jar!/locale/de-AT/prefbar/contents.rdf'
                                           );

    unreg.unregister();

    var PrefbarPath = unreg.UChrome + 'prefbar.jar';
    if (!unreg.exists(PrefbarPath)) PrefbarPath = unreg.Chrome + 'prefbar.jar';
    alert(goPrefBar.GetString("prefbarOverlayStrings", "uninstalldeletefile").replace("$filename", PrefbarPath));

    // Remove XUL.mfasl
    try {
        var FastLoadService = Components.classes["@mozilla.org/fast-load-service;1"]
            .getService(Components.interfaces.nsIFastLoadService);
        var cachefile = FastLoadService.newFastLoadFile("XUL");
        cachefile.remove(false);
    }
    catch(e) {}
}

function prefbarLogError(e, lf, id, fname) {
    // Nothing to display for us. Let the internal error management do the job
    if (!e.message) {
        throw(e);
        return;
    }

    var olnum = e.lineNumber;
    // SeaMonkey has "fileName", Firefox has "filename"
    var osrc = e.fileName ? e.fileName : e.filename;
    if (osrc.match(/prefbarOverlay\.js/)) {
        olnum = e.lineNumber - lf.lineNumber;
        osrc = "'" + fname + "' in button '" + id.substr(20) + "'";
    }
    var omsg = "PrefBar error: " + e.message;

    var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
                         .getService(Components.interfaces.nsIConsoleService);
    var scriptError = Components.classes["@mozilla.org/scripterror;1"]
                      .createInstance(Components.interfaces.nsIScriptError);
    scriptError.init(omsg, osrc, null, olnum, null, 2, null);
    consoleService.logMessage(scriptError);
}

function prefbarUpdateNotify() {
    if (goPrefBar.updateNotifyDone) return;

    var prefenabled = "extensions.prefbar.updatenotify.enabled";
    var prefscver = "extensions.prefbar.updatenotify.curversion";

    var enabled = navigator.preference(prefenabled);
    if (!enabled) return;
    var scver = navigator.preference(prefscver);
    var ourver = goPrefBar.prefbarVersion;

    // User updated PrefBar manually or initial run of update notifier
    if (ourver > scver)
        navigator.preference(prefscver, ourver);
    // User already has been notified before, so exit
    else if (scver > ourver)
        return;

    var request = new XMLHttpRequest();
    request.onload = prefbarUpdateResponse;
    request.open("GET", "http://prefbar.mozdev.org/curversion.info", true);
    request.send(null);
}
function prefbarUpdateResponse(event) {
    var response = event.target;
    if (response.status != 200) return;
    var cver = response.responseText.substr(0,8);
    if (!cver.match(/^[0-9]+$/)) return;
    cver = Number(cver);

    var prefscver = "extensions.prefbar.updatenotify.curversion";
    var scver = navigator.preference(prefscver);
    if (cver > scver) {
        navigator.preference(prefscver, cver);
        var msg = goPrefBar.GetString("prefbarOverlayStrings",
                                      "updateavailable");
        if (goPrefBar.msgYesNo(window, msg))
            prefbarGoLink("http://prefbar.mozdev.org/");
    }

    goPrefBar.updateNotifyDone = true;
}

// Just a demo function. Maybe replaced in future with real Icon support...
function prefbarSetIconDemo(button, dataurl) {
    if (button.tagName == "checkbox") {
        button.setAttribute("src", dataurl);
        button.label = "";
    }
    else if (button.tagName == "toolbarbutton") {
        button.image = dataurl;
        button.label = "";
    }
}
